import json
import shutil
from dataclasses import dataclass, field, asdict
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional

from pylizlib.core.data.gen import gen_random_string
from pylizlib.core.os.path import random_subfolder, clear_folder_contents, clear_or_move_to_temp, duplicate_directory


# S = TypeVar("S")


# class CatalogueInterface(ABC, Generic[S]):
#
#     def __init__(self, path_catalogue: Path):
#         self.__setup_catalogue(path_catalogue)
#
#     def __setup_catalogue(self, path_catalogue: Path):
#         path_catalogue.mkdir(parents=True, exist_ok=True)
#         self.path_catalogue = path_catalogue
#
#     def update_catalogue_path(self, new_path: Path):
#         self.__setup_catalogue(new_path)
#
#     @abstractmethod
#     def add(self, data: S) -> S:
#         pass
#
#     @abstractmethod
#     def get_all(self) -> list[S]:
#         pass


@dataclass
class SnapDirAssociation:
    index: int
    original_path: str
    folder_id: str
    _current_index: ClassVar[int] = 0


    @classmethod
    def next_index(cls):
        cls._current_index += 1
        return cls._current_index

    @property
    def directory_name(self) -> str:
        return self.index.__str__() + "-" + Path(self.original_path).name


    @staticmethod
    def gen_random(source_folder_for_choices: Path):
        return SnapDirAssociation(
            index=SnapDirAssociation.next_index(),
            original_path=random_subfolder(source_folder_for_choices).__str__(),
            folder_id=gen_random_string(4)
        )

    @staticmethod
    def gen_random_list(count: int, source_folder_for_choices: Path) -> list['SnapDirAssociation']:
        return [SnapDirAssociation.gen_random(source_folder_for_choices) for _ in range(count)]


    def copy_install_to(self, catalogue_target_path: Path):
        source = Path(self.original_path)
        destination = catalogue_target_path.joinpath(self.directory_name)
        destination.mkdir(parents=True, exist_ok=True)

        for src_path in source.iterdir():
            dst_path = destination / src_path.name
            if src_path.is_dir():
                shutil.copytree(src_path, dst_path)
            else:
                shutil.copy2(src_path, dst_path)


class SnapEditType(Enum):
    ADD_DIR = "Add"
    REMOVE_DIR = "Remove"


@dataclass
class SnapEditAction:
    action_type: SnapEditType
    timestamp: datetime = datetime.now()
    new_path: str = ""
    folder_id_to_remove: str = ""



@dataclass
class Snapshot:
    id: str
    name: str
    desc: str
    author: str = field(default="UnknownUser")
    directories: list[SnapDirAssociation] = field(default_factory=list)
    tags: list[str] = field(default_factory=list)
    date_created: datetime = datetime.now()
    date_modified: datetime | None = None
    date_last_used: datetime | None = None
    date_last_modified: datetime | None = None
    data: dict[str, str] = field(default_factory=dict)

    @property
    def tags_as_string(self) -> str:
        return ", ".join(self.tags) if self.tags else " "

    def get_for_table_array(self, key_list: list[str]) -> list[str]:
        array = [self.name, self.desc]
        for key in key_list:
            value = self.data.get(key, "")
            array.append(value)
        array.append(self.date_created.strftime("%d/%m/%Y %H:%M:%S"))
        array.append(self.tags_as_string)
        return array

    @property
    def folder_name(self) -> str:
        return self.id + "-" + self.name

    def add_data_item(self, key: str, value: str) -> None:
        """Aggiunge un elemento al dizionario."""
        self.data[key] = value

    def remove_data_item(self, key: str) -> Optional[str]:
        """Rimuove un elemento dal dizionario e restituisce il valore rimosso."""
        return self.data.pop(key, None)

    def has_data_item(self, key: str) -> bool:
        """Verifica se una chiave esiste nel dizionario."""
        return key in self.data

    def get_data_item(self, key: str, default: str = "") -> str:
        """Ottiene un valore dal dizionario con un default."""
        return self.data.get(key, default)

    def clear_all_data(self) -> None:
        """Pulisce tutti gli elementi del dizionario."""
        self.data.clear()

    def edit_data_item(self, key: str, new_value: str) -> None:
        """Modifica il valore di un elemento esistente nel dizionario."""
        if key in self.data:
            self.data[key] = new_value
        else:
            raise KeyError(f"Key '{key}' not found in data.")












class SnapshotUtils:

    @staticmethod
    def gen_random_snap(source_folder_for_choices: Path, id_length: int = 10, ) -> Snapshot:
        dirs = SnapDirAssociation.gen_random_list(3, source_folder_for_choices)
        return Snapshot(
            id=gen_random_string(id_length),
            name="Snapshot " + gen_random_string(5),
            desc="Randomly generated snapshot",
            author="User",
            directories=dirs,
            tags=["example", "test"]
        )


    @staticmethod
    def get_snapshot_from_path(path_snapshot: Path, json_filename: str) -> Snapshot | None:
        if path_snapshot.is_file():
            raise ValueError(f"The provided path {path_snapshot} is not a directory.")
        if not path_snapshot.exists():
            raise FileNotFoundError(f"The provided path {path_snapshot} does not exist.")
        path_snapshot_json = path_snapshot.joinpath(json_filename)
        if not path_snapshot_json.is_file():
            raise FileNotFoundError(f"No snapshot.json file found in {path_snapshot}.")
        return SnapshotSerializer.from_json(path_snapshot_json)

    @staticmethod
    def get_snapshot_path(folder_name: str, catalogue_path: Path) -> Path:
        return catalogue_path.joinpath(folder_name)

    @staticmethod
    def get_snapshot_json_path(folder_name: str, catalogue_path: Path, json_filename: str) -> Path:
        return SnapshotUtils.get_snapshot_path(folder_name, catalogue_path).joinpath(json_filename)




class SnapshotSerializer:

    @staticmethod
    def _converter(o):
        """Converti datetime e enum in formati serializzabili JSON."""
        if isinstance(o, datetime):
            return o.isoformat()
        if isinstance(o, Enum):
            return o.value  # oppure o.name se preferisci il nome
        raise TypeError(f"Object of type {type(o).__name__} is not JSON serializable")

    @staticmethod
    def to_json(snapshot: Snapshot, path: Path) -> None:
        data_dict = asdict(snapshot)
        json_str = json.dumps(data_dict, default=SnapshotSerializer._converter, indent=4)
        path.write_text(json_str, encoding="utf-8")


    @classmethod
    def from_json(cls, filepath: Path) -> Snapshot:
        """Legge un AtomDevConfig da file JSON, convertendo datetime ed enum"""
        data = json.loads(filepath.read_text(encoding="utf-8"))

        # Converte i campi datetime da stringa ISO8601 a datetime
        for key in ["date_created", "date_last_installed", "date_modified", "date_last_used", "date_last_modified"]:
            if key in data and data[key] is not None:
                data[key] = datetime.fromisoformat(data[key])

        # Conversione 'directories' in lista di ConfigDirAssociation
        if "directories" in data and isinstance(data["directories"], list):
            data["directories"] = [SnapDirAssociation(**d) if isinstance(d, dict) else d for d in data["directories"]]

        return Snapshot(**data)

    @classmethod
    def update_field(cls, filepath: Path, field_name: str, new_value):
        # Leggi dati esistenti dal file JSON
        data = json.loads(filepath.read_text(encoding="utf-8"))

        # Aggiorna solo il campo specificato
        data[field_name] = new_value

        # Serializza di nuovo il file con i convertitori per datetime e enum se necessario
        json_str = json.dumps(data, default=cls._converter, indent=4)
        filepath.write_text(json_str, encoding="utf-8")



class SnapshotManager:

    def __init__(
            self,
            snapshot: Snapshot,
            catalogue_path: Path,
            json_filename: str = "snapshot.json"
    ):
        self.snapshot = snapshot
        self.json_filename = json_filename
        self.path_catalogue = catalogue_path
        self.path_snapshot = SnapshotUtils.get_snapshot_path(self.snapshot.folder_name, self.path_catalogue)
        self.path_snapshot_json = SnapshotUtils.get_snapshot_json_path(self.snapshot.folder_name, self.path_catalogue, self.json_filename)

    def __save_json(self):
        SnapshotSerializer.to_json(self.snapshot, self.path_snapshot_json)

    def create(self):
        if self.path_snapshot.exists():
            clear_folder_contents(self.path_snapshot)
        self.path_snapshot.mkdir(parents=True, exist_ok=True)
        for snap_dir in self.snapshot.directories:
            snap_dir.copy_install_to(self.path_snapshot)
        self.__save_json()

    def delete(self):
        if self.path_snapshot.exists():
            clear_or_move_to_temp(self.path_snapshot)

    def update_json_data_fields(self):
        SnapshotSerializer.update_field(self.path_snapshot_json, "data", self.snapshot.data)
        SnapshotSerializer.update_field(self.path_snapshot_json, "date_last_modified", datetime.now().isoformat())
        self.snapshot.date_last_modified = datetime.now()

    def update_json_base_fields(self):
        SnapshotSerializer.update_field(self.path_snapshot_json, "name", self.snapshot.name)
        SnapshotSerializer.update_field(self.path_snapshot_json, "desc", self.snapshot.desc)
        SnapshotSerializer.update_field(self.path_snapshot_json, "author", self.snapshot.author)
        SnapshotSerializer.update_field(self.path_snapshot_json, "tags", self.snapshot.tags)
        SnapshotSerializer.update_field(self.path_snapshot_json, "date_modified", datetime.now().isoformat())
        self.snapshot.date_modified = datetime.now()

    def install_directory(self, destination_path: Path):
        if not destination_path.exists() or not destination_path.is_dir():
            raise ValueError(f"The provided path {destination_path} is not a valid directory.")
        new_dir = SnapDirAssociation(
            index=SnapDirAssociation.next_index(),
            original_path=destination_path.as_posix(),
            folder_id=gen_random_string(4)
        )
        new_dir.copy_install_to(self.path_snapshot)
        self.snapshot.directories.append(new_dir)
        self.__save_json()

    def uninstall_directory_by_folder_id(self, folder_id: str):
        dir_to_remove = next((d for d in self.snapshot.directories if d.folder_id == folder_id), None)
        if dir_to_remove:
            dir_path = self.path_snapshot.joinpath(dir_to_remove.directory_name)
            if dir_path.exists():
                clear_or_move_to_temp(dir_path)
            self.snapshot.directories.remove(dir_to_remove)
            self.__save_json()

    def update_from_actions_list(self, edits: list[SnapEditAction]):
        for edit in edits:
            if edit.action_type == SnapEditType.ADD_DIR:
                self.install_directory(Path(edit.new_path))
            elif edit.action_type == SnapEditType.REMOVE_DIR:
                self.uninstall_directory_by_folder_id(edit.folder_id_to_remove)

    def duplicate(self):
        if not self.path_snapshot.exists():
            raise FileNotFoundError(f"The snapshot path {self.path_snapshot} does not exist.")
        new_snap = self.snapshot
        new_snap.id = gen_random_string(10)
        new_snap.name = self.snapshot.name + " Copy"
        new_snap.date_created = datetime.now()
        new_snap_path = SnapshotUtils.get_snapshot_path(new_snap.folder_name, self.path_catalogue)
        new_snap_json_path = SnapshotUtils.get_snapshot_json_path(new_snap.folder_name, self.path_catalogue, self.json_filename)
        duplicate_directory(self.path_snapshot, new_snap_path, "")
        SnapshotSerializer.to_json(new_snap, new_snap_json_path)

    def install(self):
        for dir_assoc in self.snapshot.directories:
            dir_assoc.copy_install_to(Path(dir_assoc.original_path))
        self.snapshot.date_last_used = datetime.now()
        SnapshotSerializer.update_field(self.path_snapshot_json, "date_last_used", self.snapshot.date_last_used.isoformat())


class SnapshotCatalogue:

    def __init__(
            self,
            path_catalogue: Path,
            snapshot_json_filename: str = "snapshot.json"
    ):
        self.path_catalogue = path_catalogue
        self.snapshot_json_filename = snapshot_json_filename

        self.path_catalogue.mkdir(parents=True, exist_ok=True)


    def add(self, snap: Snapshot):
        snap_manager = SnapshotManager(snap, self.path_catalogue, self.snapshot_json_filename)
        snap_manager.create()

    def delete(self, snap: Snapshot):
        snap_manager = SnapshotManager(snap, self.path_catalogue, self.snapshot_json_filename)
        snap_manager.delete()

    def get_all(self) -> list[Snapshot]:
        snapshots: list[Snapshot] = []
        for current_dir in self.path_catalogue.iterdir():
            if current_dir.is_dir():
                snap = SnapshotUtils.get_snapshot_from_path(current_dir, self.snapshot_json_filename)
                if snap is not None:
                    snapshots.append(snap)
        return snapshots

    def get_by_id(self, snap_id: str) -> Optional[Snapshot]:
        all_snaps = self.get_all()
        for snap in all_snaps:
            if snap.id == snap_id:
                return snap
        return None

    def update_snapshot(self, snap: Snapshot, edits: list[SnapEditAction]):
        snap_manager = SnapshotManager(snap, self.path_catalogue, self.snapshot_json_filename)
        snap_manager.update_json_base_fields()
        snap_manager.update_json_data_fields()
        snap_manager.update_from_actions_list(edits)

    def duplicate_by_id(self, snap_id: str):
        snap = self.get_by_id(snap_id)
        if snap is None:
            raise ValueError(f"No snapshot found with ID {snap_id}")
        snap_manager = SnapshotManager(snap, self.path_catalogue, self.snapshot_json_filename)
        snap_manager.duplicate()

    def install(self, snap: Snapshot):
        snap_manager = SnapshotManager(snap, self.path_catalogue, self.snapshot_json_filename)
        snap_manager.install()







from dataclasses import dataclass

from PySide6.QtGui import QColor

from pylizlib.core.domain.os import OsTheme
from pylizlib.qt.domain.theme import AppTheme


@dataclass
class ResIcon:
    res_id: str
    color: QColor | None = None
    color_light: QColor | None = None
    color_dark: QColor | None = None

    def get_color(self, theme: AppTheme):
        if self.color_light is not None and self.color_dark is not None:
            match theme.system_ui_mode:
                case OsTheme.LIGHT:
                    return self.color_light
                case OsTheme.DARK:
                    return self.color_dark
                case _:
                    return self.color
        return self.color

    def has_fixed_color(self):
        return (self.color is not None) or (self.color_light is None and self.color_dark is None)


@dataclass
class ResImage:
    res_id: str



import os
from enum import Enum
from typing import List

from pylizlib.core.app.configini import IniManager, IniItem
from pylizlib.core.log.pylizLogger import logger
from pylizlib.core.os import path as pathutils

from dataclasses import dataclass


class PylizDirFoldersTemplate(Enum):
    """
    Enum for standard folder templates.
    """
    MODELS = "models"
    AI = "ai"
    TEMP = "temp"
    LOGS = "logs"
    RESULTS = "results"


@dataclass
class PylizDirFolder:
    """
    Dataclass for storing folder information.
    """
    key: str
    name: str
    path: str


class PylizApp:
    """
    Class for managing application directories and configuration.
    """

    __folders: List[PylizDirFolder] = []
    __ini: IniManager | None = None
    __ini_path: str | None = None
    __ini_initialized = False

    language_version: str | None = None


    def __init__(
            self,
            app_name: str,
            app_version: str | None = None,
            folder_name: str | None = None,
            author: str = "Unknown"
    ):
        """
        Constructor for PylizApp class.
        :param folder_name: Full name of the folder.
        :param app_name: Name of the application.
        :param app_version: Version of the application.
        :param author: Author of the application.
        """
        # Settaggio path
        app_folder_name = folder_name if folder_name is not None else app_name
        self.path: str = pathutils.get_app_home_dir(app_folder_name)
        # Cartella pyliz
        pathutils.check_path(self.path, True)
        pathutils.check_path_dir(self.path)
        # Settaggio variabili
        self.name = app_name
        self.version = app_version
        self.author = author

    def get_path(self):
        return self.path

    # FOLDERS --------------------------------------------

    def add_folder(self, key: str, folder_name: str):
        folder_path = os.path.join(self.path, folder_name)
        pathutils.create_path(folder_path)
        pathutils.check_path(folder_path, True)
        pathutils.check_path_dir(folder_path)
        self.__folders.append(PylizDirFolder(key, folder_name, folder_path))
        return folder_path

    def add_template_folder(self, template_key: PylizDirFoldersTemplate, name: str | None = None):
        folder_name = name if name is not None else template_key.value
        self.add_folder(template_key.value, folder_name)

    def add_all_template_folders(self):
        for template_key in PylizDirFoldersTemplate:
            self.add_template_folder(template_key)

    def add_folder_with_ini(self, key: str, folder_name: str, ini_section: str, ini_key: str):
        folder_path = self.add_folder(key, folder_name)
        self.set_ini_value(ini_section, ini_key, folder_path)

    def get_folder_path(self, key: str):
        for folder in self.__folders:
            if folder.key == key:
                return folder.path
        return None

    def get_folder_template_path(self, template_key: PylizDirFoldersTemplate, add_if_not_exist: bool = True):
        path = self.get_folder_path(template_key.value)
        if path is None and add_if_not_exist:
            return self.add_template_folder(template_key)
        else:
            return path

    def check_for_all_init(self):
        if not self.__ini.exists() or not self.__ini_initialized:
            raise Exception("PylizDirError: Configuration file not initialized or not found")

    # INI --------------------------------------------

    def create_ini(self, config_name: str, list_of_items: List[IniItem] | None = None):
        self.__ini_path = os.path.join(self.path, config_name)
        self.__ini = IniManager(self.__ini_path)
        if not self.__ini.exists():
            self.__ini.create(list_of_items)
        self.__ini_initialized = True

    def get_ini_value(self, section, key, is_bool=False):
        self.check_for_all_init()
        return self.__ini.read(section, key, is_bool)

    def set_ini_value(self, section, key, value):
        self.check_for_all_init()
        self.__ini.write(section, key, value)

    def get_ini_path(self):
        if self.__ini_initialized:
            return self.__ini_path
        else:
            return None

    def delete_ini(self):
        if self.__ini_initialized and os.path.exists(self.__ini_path):
            os.remove(self.__ini_path)
            self.__ini_initialized = False
            self.__ini = None
            self.__ini_path = None
        else:
            logger.warning("INI file not initialized. Nothing to delete.")




import json
from json import JSONDecodeError


class JsonUtils:

    @staticmethod
    def is_valid_json(json_str):
        try:
            json.loads(json_str)
            return True
        except JSONDecodeError:
            return False

    @staticmethod
    def has_keys(json_str, keys: list[str]) -> bool:
        try:
            json_obj = json.loads(json_str)
            for key in keys:
                if key not in json_obj:
                    return False
            return True
        except JSONDecodeError:
            return False

    @staticmethod
    def clean_json_apici(json_string):
        # Rimuovi il prefisso '''json
        if json_string.startswith("```json"):
            json_string = json_string[len("```json"):]
        # Rimuovi il suffisso ''' se presente
        if json_string.endswith("```"):
            json_string = json_string[:-len("```")]
        # Elimina eventuali spazi bianchi in eccesso
        json_string = json_string.strip()
        # Carica il JSON
        return json_string




import time
from abc import abstractmethod
from time import sleep
from typing import Callable, Any

from PySide6.QtCore import QRunnable, QObject, Signal

from pylizlib.core.data.gen import gen_random_string
from pylizlib.core.handler.progress import QueueProgress, QueueProgressMode, get_step_progress_percentage
from pylizlib.core.log.pylizLogger import logger
from pylizlib.qt.handler.operation_domain import OperationStatus, OperationInfo


class Task(QObject):
    task_update_status = Signal(str, OperationStatus)
    task_update_progress = Signal(str, int)

    def __init__(
            self,
            name: str,
            abort_all_on_error: bool = True,
    ):
        super().__init__()
        self.id = gen_random_string(10)
        self.name = name
        self.abort_all_on_error = abort_all_on_error
        self.status = OperationStatus.Pending
        self.on_progress_changed = None
        self.result: Any = None
        self.progress = 0

    def execute(self):
        return None

    def update_task_status(self, status: OperationStatus):
        logger.debug("Updating task \"%s\" status: %s", self.name, status)
        self.status = status
        self.task_update_status.emit(self.name, status)

    def update_task_progress(self, progress: int):
        logger.debug("Updating task \"%s\" progress: %s", self.name, progress)
        self.progress = progress
        self.task_update_progress.emit(self.name, progress)
        if self.on_progress_changed:
            self.on_progress_changed(self.name, progress)

    def gen_update_task_progress(self, current: int, total: int):
        self.update_task_progress(get_step_progress_percentage(current, total))


class OperationSignals(QObject):
    op_start = Signal()
    op_update = Signal(object)
    op_update_status = Signal(str, OperationStatus)
    op_update_progress = Signal(str, int)
    op_eta_update = Signal(str, str)
    op_failed = Signal(str, str)
    op_finished = Signal(object)

    task_start = Signal(str)
    task_update_status = Signal(str, OperationStatus)
    task_update_progress = Signal(str, int)
    task_failed = Signal(str, str)
    task_finished = Signal(str)


class Operation(QRunnable):

    def __init__(
            self,
            tasks: list[Task],
            op_info: OperationInfo,
    ):
        super().__init__()
        self.signals = OperationSignals()
        self.id = gen_random_string(10)
        self.info = op_info
        self.status = OperationStatus.Pending

        self.tasks = tasks
        self.progress = 0
        self.running = False
        self.error = None
        self.progress_obj = QueueProgress(QueueProgressMode.SINGLE, len(tasks))
        self.finished_callback: Callable | None = None
        self.op_progress_update_callback: Callable | None = None
        self.current_task: Task | None = None

        self.time_started = None
        self.time_elapsed = 0
        self.time_finished = None
        self.time_estimated_total = 0
        self.time_estimated_remaining = 0

        for task in tasks:
            self.progress_obj.add_single(task.name)
            task.task_update_status.connect(self.signals.task_update_status)
            task.task_update_progress.connect(self.signals.task_update_progress)

    def execute_tasks(self):
        for task in self.tasks:
            try:
                task.on_progress_changed = self.on_task_progress_update
                self.signals.task_start.emit(task.name)
                task.update_task_status(OperationStatus.InProgress)
                logger.debug("Executing task: %s", task.name)
                self.current_task = task
                result = task.execute()
                task.result = result
                task.update_task_status(OperationStatus.Completed)
            except Exception as e:
                task.update_task_status(OperationStatus.Failed)
                logger.error("Error in task %s: %s", task.name, e)
                self.signals.task_failed.emit(task.name, str(e))
                if task.abort_all_on_error:
                    raise RuntimeError(f"Task {task.name} failed: {e}")
            finally:
                self.signals.task_finished.emit(task.name)
                self.current_task = None
                sleep(self.info.delay_each_task)

    def execute(self):
        try:
            self.set_operation_started()
            self.update_op_status(OperationStatus.InProgress)
            self.execute_tasks()
            self.update_op_status(OperationStatus.Completed)
        except Exception as e:
            self.update_op_status(OperationStatus.Failed)
            self.signals.op_failed.emit(self.id, str(e))
            self.error = str(e)
            logger.error("Error in operation: %s", e)
        finally:
            self.set_operation_finished()

    def on_task_progress_update(self, task_name: str, progress: int):
        self.progress_obj.set_single_progress(task_name, progress)
        self.update_op_progress(self.progress_obj.get_total_progress())

    def run(self, /):
        self.execute()

    @abstractmethod
    def stop(self):
        pass

    def get_tasks_ids(self) -> list[str]:
        return [task.name for task in self.tasks]

    def get_task_results(self) -> list[Any]:
        return [task.result for task in self.tasks]

    def get_task_result_by_name(self, task_name: str) -> Any:
        for task in self.tasks:
            if task.name == task_name:
                return task.result
        return None

    def get_task_result_by_id(self, task_id: str) -> Any:
        for task in self.tasks:
            if task.id == task_id:
                return task.result
        return None

    def update_op_status(self, status: OperationStatus):
        logger.debug("Updating operation status: %s", status)
        self.status = status
        self.__update_times()
        self.signals.op_update_status.emit(self.id, status)
        self.signals.op_update.emit(self)

    def update_op_progress(self, progress: int):
        logger.debug("Updating operation progress: %s", progress)
        self.progress = progress
        self.__update_times()
        self.signals.op_update_progress.emit(self.id, progress)
        if self.op_progress_update_callback:
            self.op_progress_update_callback(self.id, self.progress)
        self.signals.op_update.emit(self)

    def __update_times(self):
        if self.running:
            self.time_elapsed = time.perf_counter() - self.time_started
            if self.progress > 0:
                self.time_estimated_total = self.time_elapsed / (self.progress / 100)
                self.time_estimated_remaining = max(0, self.time_estimated_total - self.time_elapsed)
                self.signals.op_eta_update.emit(self.id, self.get_eta_formatted())
        else:
            if self.time_started and self.time_finished:
                self.time_elapsed = self.time_finished - self.time_started

    def set_finished_callback(self, callback: Callable):
        self.finished_callback = callback

    def set_op_progress_callback(self, callback: Callable):
        self.op_progress_update_callback = callback

    def set_operation_started(self):
        logger.info("Starting operation %s", self.id)
        self.running = True
        self.signals.op_start.emit()
        self.time_started = time.perf_counter()
        self.__update_times()

    def set_operation_finished(self):
        logger.info("Finishing operation %s", self.id)
        self.running = False
        self.time_finished = time.perf_counter()
        self.__update_times()
        if self.finished_callback:
            self.finished_callback()
        self.signals.op_finished.emit(self)

    def get_elapsed_formatted(self) -> str:
        """Restituisce il tempo trascorso nel formato mm:ss"""
        minutes = int(self.time_elapsed) // 60
        seconds = int(self.time_elapsed) % 60
        return f"{minutes:02.0f}:{seconds:02.0f}"

    def get_eta_formatted(self) -> str:
        """
        Restituisce il tempo stimato rimanente nel formato mm:ss.
        """
        if self.progress <= 0:
            return "--:--"

        minutes = int(self.time_estimated_remaining) // 60
        seconds = int(self.time_estimated_remaining) % 60
        return f"{minutes:02}:{seconds:02}"

    def is_failed(self):
        return self.status == OperationStatus.Failed

    def is_in_progress(self):
        return self.status == OperationStatus.InProgress

    def is_completed(self):
        return self.status == OperationStatus.Completed

    def is_pending(self):
        return self.status == OperationStatus.Pending


from qfluentwidgets.common import ConfigItem


class QtFwQConfigItem(ConfigItem):

    def __init__(self, enabled: bool, group, name, default, validator=None, serializer=None, restart=False):
        super().__init__(group, name, default, validator, serializer, restart)
        self.enabled = enabled



from typing import Any, Callable

from PySide6.QtWidgets import QGroupBox, QFormLayout, QSizePolicy

from pylizlib.core.app.pylizapp import PylizApp
from pylizlib.qt.domain.preference import PreferenceConfigData, PreferenceTypes


# def create_prf_path(item: ConfigIniItem, app: PylizApp) -> PreferenceItemPath:
#     return PreferenceItemPath(
#         label=QLabel(item.name),
#         widget=PathLineSelector(),
#         config_id=item.id,
#         getter=lambda: ConfigHandler.read(item, app=app),
#         setter=lambda value: ConfigHandler.write(item, value, app=app),
#     )
#
#
# def create_prf_checkbox(item: ConfigItem, app: PylizApp) -> PreferenceItemCheck:
#     widget = QCheckBox()
#     return PreferenceItemCheck(
#         label=QLabel(item.name),
#         widget=widget,
#         config_id=item.id,
#         getter=lambda: ConfigHandler.read(item, app=app),
#         setter=lambda value: ConfigHandler.write(item, value, app=app),
#     )
#
#
# def create_prf_combo_box(item: ConfigItem, app: PylizApp) -> PreferenceItemCombo:
#     widget = QComboBox()
#     for value in item.values:
#         widget.addItem(value)
#     return PreferenceItemCombo(
#         label=QLabel(item.name),
#         widget=widget,
#         config_id=item.id,
#         getter=lambda: ConfigHandler.read(item, app=app),
#         setter=lambda value: ConfigHandler.write(item, value, app=app),
#     )
#
#
# def create_prf_spin_box(item: ConfigItem, app: PylizApp) -> PreferenceItemSpinBox:
#     widget = QSpinBox()
#     widget.setMinimum(int(item.min_value))
#     widget.setMaximum(int(item.max_value))
#     widget.setSingleStep(1)
#     return PreferenceItemSpinBox(
#         label=QLabel(item.name),
#         widget=widget,
#         config_id=item.id,
#         getter=lambda: int(ConfigHandler.read(item, app=app)),
#         setter=lambda value: ConfigHandler.write(item, str(value), app=app),
#     )


def create_prf_group_box_form_layout(name: str, parent: Any) -> tuple:
    group_box = QGroupBox(name, parent)
    group_box.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
    form_layout = QFormLayout(group_box)
    form_layout.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
    group_box.setLayout(form_layout)
    return group_box, form_layout


# def add_ui_to_form_layout(
#         data: PreferenceConfigData,
#         layout: QFormLayout,
#         app: PylizApp,
#         on_change: Callable[[ConfigIniItem, str | None, bool | None], None],
# ) -> None:
#     match data.type:
#         case PreferenceTypes.PATH:
#             element = create_prf_path(data.config, app)
#             element.widget.setText(element.getter())
#             element.widget.get_text_changed().connect(element.setter)
#             element.widget.get_text_changed().connect(lambda text: on_change(data.config, text, None))
#             layout.addRow(element.label, element.widget)
#         case PreferenceTypes.CHECK:
#             element = create_prf_checkbox(data.config, app)
#             element.widget.setChecked(element.getter())
#             element.widget.stateChanged.connect(lambda state: element.setter(bool(state)))
#             element.widget.stateChanged.connect(lambda state: on_change(data.config, None, bool(state)))
#             layout.addRow(element.label, element.widget)
#             pass
#         case PreferenceTypes.COMBO:
#             element = create_prf_combo_box(data.config, app)
#             element.widget.setCurrentText(element.getter())
#             element.widget.currentTextChanged.connect(element.setter)
#             element.widget.currentTextChanged.connect(lambda text: on_change(data.config, text, None))
#             layout.addRow(element.label, element.widget)
#         case PreferenceTypes.SPINBOX:
#             element = create_prf_spin_box(data.config, app)
#             element.widget.setValue(element.getter())
#             element.widget.valueChanged.connect(element.setter)
#             element.widget.valueChanged.connect(lambda value: on_change(data.config, str(value), None))
#             layout.addRow(element.label, element.widget)


from PySide6.QtWidgets import QHBoxLayout, QLineEdit, QPushButton, QFileDialog, QWidget


class PathLineSelector(QWidget):

    def __init__(self, parent=None, select_file=False):
        super().__init__(parent)

        self.select_file = select_file  # True per selezionare file, False per cartelle

        self.layout = QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)  # Rimuove i margini

        self.line_edit = QLineEdit(self)
        self.button = QPushButton("...", self)

        self.layout.addWidget(self.line_edit)
        self.layout.addWidget(self.button)

        self.button.clicked.connect(self.open_dialog)

    def open_dialog(self):
        if self.select_file:
            path, _ = QFileDialog.getOpenFileName(self, "Select File")
        else:
            path = QFileDialog.getExistingDirectory(self, "Select Directory")

        if path:
            self.line_edit.setText(path)

    def text(self):
        return self.line_edit.text()

    def setText(self, text):
        self.line_edit.setText(text)

    def get_text_changed(self):
        return self.line_edit.textChanged




import subprocess


def is_windows_dark_theme() -> bool:
    import platform
    if platform.system() != "Windows":
        return False

    try:
        import winreg
        key = winreg.OpenKey(
            winreg.HKEY_CURRENT_USER,
            r"Software\Microsoft\Windows\CurrentVersion\Themes\Personalize"
        )
        value, _ = winreg.QueryValueEx(key, "AppsUseLightTheme")
        return value == 0  # 0 = dark, 1 = light
    except Exception:
        return False


def is_macos_dark_theme():
    try:
        result = subprocess.run(
            ["defaults", "read", "-g", "AppleInterfaceStyle"],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True
        )
        return result.stdout.strip().lower() == "dark"
    except Exception:
        return False


def is_dark_theme() -> bool:
    try:
        # Windows specific check
        import platform
        if platform.system() == "Windows":
            return is_windows_dark_theme()
        elif platform.system() == "Darwin":
            return is_macos_dark_theme()
        else:
            # Fallback cross-platform using Qt palette
            raise Exception("Unsupported OS")
    except Exception:
        raise Exception("An error occurred while checking the theme. Please check your system settings.")


from PySide6.QtCore import Signal, QSize, Qt
from PySide6.QtWidgets import QWidget, QHBoxLayout, QSizePolicy
from qfluentwidgets import BodyLabel, ToolButton, FluentIcon


class FileItem(QWidget):
    """ File item """

    removed = Signal(QWidget)

    def __init__(self, file: str, parent=None):
        super().__init__(parent=parent)
        self.file = file
        self.hBoxLayout = QHBoxLayout(self)
        self.fileLabel = BodyLabel(file, self)
        self.removeButton = ToolButton(FluentIcon.CLOSE, self)

        self.removeButton.setFixedSize(39, 29)
        self.removeButton.setIconSize(QSize(12, 12))

        self.setFixedHeight(53)
        self.setSizePolicy(QSizePolicy.Policy.Ignored, QSizePolicy.Policy.Fixed)
        self.hBoxLayout.setContentsMargins(48, 0, 60, 0)
        self.hBoxLayout.addWidget(self.fileLabel, 0, Qt.AlignmentFlag.AlignLeft)
        self.hBoxLayout.addSpacing(16)
        self.hBoxLayout.addStretch(1)
        self.hBoxLayout.addWidget(self.removeButton, 0, Qt.AlignmentFlag.AlignRight)
        self.hBoxLayout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        self.removeButton.clicked.connect(lambda: self.removed.emit(self))




import getpass
import os
import shutil
import subprocess
import platform
from pathlib import Path

import psutil


PATH_DEFAULT_GIT_BASH = Path(r"C:\Program Files\Git\bin\bash.exe")


def get_folder_size_mb(path) -> float:
    """
    Get the size of a folder in megabytes
    :param path: path to the folder
    :return: size of the folder in megabytes
    """
    # Inizializza la dimensione totale a 0
    total_size = 0
    # Scansione delle cartelle e dei file all'interno del percorso dato
    for root, dirs, files in os.walk(path):
        # Aggiungi le dimensioni dei file alla dimensione totale
        for file in files:
            file_path = os.path.join(root, file)
            total_size += os.path.getsize(file_path)
    # Converti la dimensione totale in megabyte (MB)
    total_size_mb = total_size / (1024 * 1024)
    return total_size_mb


def open_system_folder(path):
    """
    Open a system folder in the default file explorer
    :param path: path to the folder
    :return:
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"Path {path} does not exist!")
    if os.name == 'nt':  # For Windows
        subprocess.Popen(['explorer', path])
    elif os.name == 'posix':  # For Linux, Mac
        subprocess.Popen(['open', path])
    else:
        raise OSError("Unsupported OS")


def has_disk_free_space(path_of_disk, mb_free):
    """
    Check if a disk has enough free space in megabytes
    :param path_of_disk: The path of the disk to check
    :param mb_free: The minimum amount of free space in megabytes
    :return: True if the disk has enough free space, False otherwise
    """
    stat = shutil.disk_usage(path_of_disk)
    spazio_disponibile_mb = stat.free / (1024 * 1024)
    if spazio_disponibile_mb > mb_free:
        return True
    else:
        return False


def get_free_space_mb(directory) -> float:
    """
    Get the free space in megabytes of a directory
    :param directory: path to the directory
    :return: free space in megabytes
    """
    statvfs = os.statvfs(directory)
    # Calculate the free space in bytes and convert to megabytes
    free_space = statvfs.f_frsize * statvfs.f_bavail / (1024 * 1024)
    return free_space


def get_directory_size(path) -> float:
    """
    Get the size of a directory in megabytes
    :param path: path to the directory
    :return: size of the directory in megabytes
    """
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size / (1024 * 1024)


def check_move_dirs_free_space(src_path, dst_path) -> bool:
    """
    Check if there is enough free space in the destination directory to move the source directory
    :param src_path: The path of the source directory
    :param dst_path: The path of the destination directory
    :return: True if there is enough free space, False otherwise
    """
    # Calculate the size of the source directory
    src_size_mb = get_directory_size(src_path)
    # Get the free space of the destination directory
    free_space_mb = get_free_space_mb(dst_path)
    # Check if there is enough space
    return free_space_mb >= src_size_mb


def is_command_available_with_run(command: str) -> bool:
    """
    Check if a command is available in the system
    :param command: The command to check
    :return: True if the command is available, False otherwise
    """
    try:
        subprocess.run([command], check=True)
        return True
    except FileNotFoundError:
        return False


def is_command_available(command: str) -> bool:
    return shutil.which(command) is not None


def is_os_unix() -> bool:
    """
    Check if the operating system is Unix-based (Linux, MacOS)
    :return: True if the operating system is Unix-based, False otherwise
    """
    current_os = platform.system()
    return current_os in ["Linux", "Darwin"]

def is_os_windows() -> bool:
    """
    Check if the operating system is Windows
    :return: True if the operating system is Windows, False otherwise
    """
    current_os = platform.system()
    return current_os == "Windows"

def is_software_installed(exe_path: Path) -> bool:
    """
    Check if a software is installed by checking if the executable file exists
    :param exe_path: The path of the executable file
    :return: True if the software is installed, False otherwise
    """
    return os.path.isfile(exe_path.__str__()) and os.access(exe_path.__str__(), os.X_OK)


def get_system_username() -> str:
    """
    Get the current system username
    :return: The current system username
    """
    return getpass.getuser()


class WindowsOsUtils:

    @staticmethod
    def is_exe_running(exe_path: Path) -> bool:
        """
        Check if an executable is running
        :param exe_path: The path of the executable file
        :return: True if the executable is running, False otherwise
        """
        exe_path_str = os.path.abspath(exe_path.__str__())  # normalize
        for proc in psutil.process_iter(['exe', 'name']):
            try:
                if proc.info['exe'] and os.path.abspath(proc.info['exe']) == exe_path_str:
                    return True
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        return False

    @staticmethod
    def get_windows_exe_version(exe_path: Path) -> str:
        """
        Get the version of a Windows executable
        if the version information is not available, returns "N/A"
        if the OS is not Windows, returns N/A
        :param exe_path: the path of the executable file
        :return:  The version of the executable or "N/A" if not available
        """
        if not is_os_windows():
            return "N/A"
        import win32api
        try:
            info = win32api.GetFileVersionInfo(exe_path.__str__(), "\\")
            # Le chiavi di versione sono memorizzate come tuple
            # Prima otteniamo la lingua e il codice di pagina
            lang, codepage = win32api.GetFileVersionInfo(exe_path.__str__(), "\\VarFileInfo\\Translation")[0]
            str_info_path = f"\\StringFileInfo\\{lang:04X}{codepage:04X}\\ProductVersion"
            version = win32api.GetFileVersionInfo(exe_path.__str__(), str_info_path)
            return version
        except Exception:
            return "N/A"

    def get_service_executable_path(service_name: str) -> str | None:
        """
        Get the executable path of a Windows service
        If the os is not Windows, returns None
        :param service_name: The name of the service
        :return: The executable path of the service or None if not found
        """
        if not is_os_windows():
            return None
        import winreg
        try:
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                 rf"SYSTEM\CurrentControlSet\Services\{service_name}")
            image_path, _ = winreg.QueryValueEx(key, "ImagePath")
            winreg.CloseKey(key)
            # Rimuove eventuali argomenti e prende solo il percorso
            image_path = image_path.strip()
            if image_path.startswith('"'):
                # Prende il contenuto tra le virgolette
                image_path = image_path.split('"')[1]
            else:
                # Prende la prima parte fino al primo spazio (se non ci sono virgolette)
                parts = image_path.split(' ')
                # Ricostruisce il percorso se contiene spazi e termina con .exe
                exe_parts = []
                for part in parts:
                    exe_parts.append(part)
                    if part.lower().endswith('.exe'):
                        break
                image_path = ' '.join(exe_parts)
            return image_path
        except Exception:
            return None


    def is_service_running(service_name: str) -> bool:
        """
        Check if a Windows service is running
        :param service_name: The name of the service
        :return: True if the service is running, False otherwise
        """
        import win32service
        import win32serviceutil
        try:
            status = win32serviceutil.QueryServiceStatus(service_name)[1]
            return status == win32service.SERVICE_RUNNING
        except Exception:
            return False

    def get_service_version(service_name: str) -> str | None:
        """
        Get the version of a Windows service by checking the version of its executable
        :param service_name: The name of the service
        :return: The version of the service or None if not found
        """
        import win32api
        path = WindowsOsUtils.get_service_executable_path(service_name)
        if path is None:
            return None
        try:
            # Funzione per ottenere versione dal file binario
            info = win32api.GetFileVersionInfo(path, '\\')
            ms = info['FileVersionMS']
            ls = info['FileVersionLS']
            version = f"{ms >> 16}.{ms & 0xFFFF}.{ls >> 16}.{ls & 0xFFFF}"
            return version
        except Exception:
            return "N/A"


from dataclasses import dataclass, field

from PySide6.QtGui import QColor

from pylizlib.core.data.gen import gen_random_string
from pylizlib.core.domain.os import OsTheme


@dataclass
class UiTheme:
    id: str = gen_random_string(10)
    mode: OsTheme | None = None
    primary_color_light: QColor | None = None
    primary_color_dark: QColor | None = None
    primary_color: QColor | None = None


    def get_primary_color(self) -> QColor | None:
        if self.mode is None:
            return self.primary_color
        else:
            return self.primary_color_light if self.mode == OsTheme.LIGHT else self.primary_color_dark


@dataclass
class AppTheme:
    ui: UiTheme = field(default_factory=UiTheme)
    system_ui_mode: OsTheme = OsTheme.UNKNOWN
